#include <u.h>
#include <libc.h>
#include <bio.h>

/* Remove clashes */
#undef warn
#define warn protowarn
#define getmode proto_getmode

static int notdot();
static void *emalloc(long);
static void *erealloc(void*, long);
static char *estrdup(char*);
static char *estrstrdup(char*, char*);
static char *estrpath(char*, char*);

static int ispathprefix(char*, char*);
//static int localkids(char*, Dir**, int);

static int proto_getfields(char*, char**, int, int, char*);
static int proto_tokenize(char*, char**, int);

enum {
	TABSTOP = 8,		/* tab counts as this many spaces */
	HUNKS = 128,		/* dir entries at a time */
};

/* random crap to pass around */
typedef struct Aux Aux;
struct Aux {
	Protoenum *penum;
	Protowarn *warn;
	char *proto;
	Biobuf *b;

	int lineno;
	char *error;

	void *arg;
};

/* a parsed proto file line */
typedef struct Line Line;
struct Line {
	int indent;
	char *new;
	char *elem;
	ulong mode;
	char *uid;
	char *gid;
	char *old;
};

static void
warn(Aux *aux, char *fmt, ...)
{
	Fmt f;
	char *buf;
	va_list va;

	fmtstrinit(&f);
	va_start(va, fmt);
	if(aux->proto != nil)
		fmtprint(&f, "%s:%d: ", aux->proto, aux->lineno);
	fmtvprint(&f, fmt, va);
	va_end(va);

	buf = fmtstrflush(&f);
	aux->warn(buf, aux->arg);
	free(buf);
}

static void
mywarn(char *buf, void *unused)
{
	USED(unused); /* gcc */
	fprint(2, "warning: %s\n", buf);
}

/*
 * parse a mode: /(d|a?l?)[0-7][0-7][0-7]/
 * return ~0 on bad modes.
 */
static ulong
getmode(Aux *aux, char *p)
{
	char *s;
	ulong m;

	m = 0;
	s = p;
	if(*s == 'd'){
		m |= DMDIR;
		s++;
	}
	if(*s == 'a'){
		m |= DMAPPEND;
		s++;
	}
	if(*s == 'l'){
		m |= DMEXCL;
		s++;
	}
	if(s[0] < '0' || s[0] > '7'
	|| s[1] < '0' || s[1] > '7'
	|| s[2] < '0' || s[2] > '7'
	|| s[3]){
		warn(aux, "bad mode specification '%s'", p);
		return ~0;
	}
	m |= strtoul(s, 0, 8);
	return m;
}

/*
 * fetch a proto line, skipping blank lines and comments.
 */
static int
rdprotoline(Aux *aux, Line *line, int minindent)
{
	char *f[6], *p, *q;
	int c, nf, indent;

	/*
	 * parse:
	 *	# comment
	 * 	<blank line>
	 *	<indentation>file [mode [owner [group [sourcefile]]]]
	 */
	for(;;){
		if((q = Brdline(aux->b, '\n')) == nil)
			return -1;

		aux->lineno++;
		p = q;
		indent = 0;
		for(; (c = *p) != '\n'; p++){
			if(c == ' ')
				indent++;
			else if(c == '\t'){
				indent += TABSTOP;
				indent -= indent % TABSTOP;
			}else
				break;
		}
		if(c == '\n' || c == '#')
			continue;

		if(indent < minindent){
			Bseek(aux->b, -Blinelen(aux->b), 1);
			aux->lineno--;
			return -1;
		}

		q[Blinelen(aux->b)-1] = '\0';
		nf = proto_tokenize(p, f, nelem(f));
		while(nf < 5)
			f[nf++] = nil;

		if(f[0][0]=='$' && (p=getenv(f[0]+1)) != nil)	/* LEAK */
			line->new = p;
		else
			line->new = f[0];
		if((line->elem = utfrrune(line->new, '/')))
			line->elem++;
		else
			line->elem = line->new;
		if(f[1]==nil || strcmp(f[1], "-")==0)
			line->mode = ~0;
		else
			line->mode = getmode(aux, f[1]);
		line->uid = f[2]==nil || strcmp(f[2], "-")==0 ? nil : f[2];
		line->gid = f[3]==nil || strcmp(f[3], "-")==0 ? nil : f[3];
		line->old = f[4]==nil || strcmp(f[4], "-")==0 ? nil : f[4];
		line->indent = indent;
		return 0;
	}
}

/*
 * skip all the files in proto that could be in the 
 * current dir (that have more indentation).
 */
#if NOTDEF /* defined, not used */
static void
skipdir(Aux *aux, int indent)
{
	Line line;

	while(rdprotoline(aux, &line, indent+1) == 0)
		;
}
#endif

static Proto*
_parseproto(Aux *aux, Proto *p, int level)
{
	int first;
	Line line;
	Proto np;

	first = 1;
	while(rdprotoline(aux, &line, level) == 0){
		memset(&np, 0, sizeof np);
		if(first && ((line.elem[0]=='+' || line.elem[0]=='*') && line.elem[1]=='\0')){
			np.type = line.elem[0];
			if(line.old)
				warn(aux, "ignoring 'old' argument '%s' to %c", line.old, line.elem[0]);
			if(line.mode != ~0)
				warn(aux, "ignoring mode argument to %c", line.old, line.elem[0]);
			if(line.uid)
				np.uid = estrdup(line.uid);
			if(line.gid)
				np.gid = estrdup(line.gid);
		}else{
			np.type = 0;
			np.new = estrpath(p->new, line.elem);
			if(line.old)
				np.old = estrdup(line.old);
			else
				np.old = estrpath(p->old, line.elem);
			np.mode = line.mode;
			if(line.uid)
				np.uid = estrdup(line.uid);
			if(line.gid)
				np.gid = estrdup(line.gid);
			_parseproto(aux, &np, line.indent+1);
			np.elem = strrchr(np.new, '/')+1;
		}
		if(p->nkid%8==0)
			p->kid = erealloc(p->kid, (p->nkid+8)*sizeof(p->kid[0]));
		p->kid[p->nkid] = np;
		p->nkid++;
		first = 0;
	}
	return p;
}

Proto*
parseproto(char *proto, char *root)
{
	Aux aux;
	Proto *p;

	p = emalloc(sizeof(*p));
	p->new = estrdup("/");
	p->old = estrdup(root);
	p->uid = nil;
	p->gid = nil;
	p->mode = ~0;

	memset(&aux, 0, sizeof aux);
	aux.warn = mywarn;
	aux.proto = cleanname(proto);

	if((aux.b = Bopen(proto, OREAD)) == nil){
		werrstr("open '%s': %r", proto);
		return nil;
	}
	if(_parseproto(&aux, p, 0) == nil){
		free(p);
		return nil;
	}
	return p;
}

static void
_freeproto(Proto *p)
{
	int i;

	free(p->new);
	free(p->old);
	free(p->uid);
	free(p->gid);
	for(i=0; i<p->nkid; i++)
		_freeproto(p->kid+i);
	free(p->kid);
}

void
freeproto(Proto *p)
{
	_freeproto(p);
	free(p);
}

/*
 * if we stat a kernel device, set the user and
 * group to be sys rather than the current 
 * value of eve.
 */
static void
plan9hack(Dir *d)
{
	if(d->type != 'M'){
		if(d->uid)
			free(d->uid);
		if(d->gid)
			free(d->gid);
		d->uid = estrdup("sys");
		d->gid = estrdup("sys");
		d->mode |= ((d->mode>>3)&7) | ((d->mode>>6)&7);
	}
}

static int
enumfile(Aux *aux, char *new, ulong mode, char *uid, char *gid, char *old)
{
	Dir *d;

	if((d=dirstat(old)) == nil){
		warn(aux, "stat '%s': %r", old);
		return -1;
	}else
		plan9hack(d);

	if(mode != ~0)
		d->mode = mode;
/*
	if(uid)
		strecpy(d.uid, d.uid+NAMELEN, uid);
	if(gid)
		strecpy(d.gid, d.gid+NAMELEN, gid);
*/	
	return aux->penum(new, old, d, aux->arg);
}

static int
execprotodir(Aux *aux, char *new, char *uid, char *gid, char *old, int type)
{
	char *newkid, *oldkid;
	int i, n, len;
	Dir *d;

	struct dirent **dentp;
	char *s;

	dentp = nil;
	s = nil;
	n = 0;

	if((d=dirstat(old)) == nil || (d->mode&DMDIR)==0){
		warn(aux, "'%s' not a directory", old);
		return -1;
	}

	n = scandir(old, &dentp, notdot, alphasort);
	len = strlen(old)+1+sizeof(dentp[i]->d_name)+1;
	s = emalloc(len);
	for(i=0; i<n; i++){
		/* Avoid sprint since it may think it is reading UTF-8 */
		s[0] = '\0';
		strcat(s, old);
		strcat(s, "/");
		strcat(s, dentp[i]->d_name);
		if((d=dirstat(s)) == nil){
			warn(aux, "cannot stat '%s'", s);
			goto Fail;
		}
		newkid = estrpath(new, dentp[i]->d_name);
		oldkid = estrpath(old, dentp[i]->d_name);
		if(d->mode&DMDIR){
			if(type=='+'){
				if(!enumfile(aux, newkid,~0, uid, gid,oldkid))
					execprotodir(aux, newkid, uid, gid, 
						     oldkid, type);
			}else
				enumfile(aux, newkid, ~0, uid, gid, oldkid);
		}else
			enumfile(aux, newkid, ~0, uid, gid, oldkid);
		free(oldkid);
		free(newkid);
	}

	if(s) 
		free(s);
	if(dentp){
		for(i=0; i<n; i++)
			free(dentp[i]);
		free(dentp);
	}

	return 0;	

 Fail:
	if(s) 
		free(s);
	if(dentp){
		for(i=0; i<n; i++)
			free(dentp[i]);
		free(dentp);
	}

	return -1;
}

static void _execproto(Aux*, Proto*);

static void
__execproto(Aux *aux, Proto *p)
{
	int i;

	for(i=0; i<p->nkid; i++){
		switch(p->kid[i].type){
		case 0:
			_execproto(aux, &p->kid[i]);
			break;
		case '*':
		case '+':
			execprotodir(aux, p->new, p->kid[i].uid, p->kid[i].gid, p->old, p->kid[i].type);
			break;
		}
	}
	return;
}

static void
_execproto(Aux *aux, Proto *p)
{

	if(enumfile(aux, p->new, p->mode, p->uid, p->gid, p->old) < 0)
		return;
	__execproto(aux, p);
	return;
}

void
execproto(Proto *p, Protoenum *penum, Protowarn *warn, void *arg)
{
	Aux aux;

	memset(&aux, 0, sizeof aux);
	aux.penum = penum;
	if(warn == nil)
		aux.warn = mywarn;
	else
		aux.warn = warn;
	aux.arg = arg;

	__execproto(&aux, p);
}

Proto*
walkproto(Proto *p, char *elem)
{
	int i;

	for(i=0; i<p->nkid; i++)
		if(p->kid[i].type==0 && ispathprefix(p->kid[i].elem, elem))
			return &p->kid[i];
	return nil;
}

char*
prototranslate(Proto *proto, char *name)
{
	char *q, *elem;
	Proto *n;

	if(name[0] != '/'){
		werrstr("bad path");
		return nil;
	}

	for(elem=name+1; elem && *elem; elem=q){
		if((q = strchr(elem, '/')))
			q++;

		if((n = walkproto(proto, elem)) == nil){
			if(proto->nkid > 0){
				switch(proto->kid[0].type){
				case '*':
					if(strchr(elem, '/'))
						return nil;
					/* fall through */
				case '+':
					return estrpath(proto->old, elem);
				}
			}
			return nil;
		}
		proto = n;
	}

	return estrdup(proto->old);
}

int
rdproto(char *pfile, char *proot, Protoenum *penm, Protowarn *pwarn, void *a)
{
	Proto *p;

	if((p = parseproto(pfile, proot)) == nil)
		return -1;

	execproto(p, penm, pwarn, a);

	freeproto(p);

	return 0;
}

static int
notdot(struct dirent *dp)
{
	char *s;

	s = dp->d_name;

	/* Not "", ".", or ".." */
	if(!s[0] || (s[0] == '.' && (!s[1] || (s[1] == '.' && !s[2]))))
		return 0;

	return 1;
}

#if 0
static int
localkids(char *path, Dir **kidsp, int incdir)
{
	int i, j, nkids;
	Dir d;
	Dir *kids;
	struct dirent **dentp;
	char *s;

	memset(&d, 0, sizeof(Dir));
	dentp = nil;
	s = nil;
	kids = nil;

	if((dirstat(path, &d) < 0) | !(d.mode&CHDIR))
		return -1;

	/* Eliminate ".", "..", and possibly all directories */
	nkids = scandir(path, &dentp, notdot, alphasort);
	kids = emalloc((nkids+1)*sizeof(Dir));
	s = emalloc(strlen(path)+NAMELEN+2);

	for(i=j=0; i<nkids; i++) {
		sprint(s, "%s/%s", path, dentp[i]->d_name);
		if(dirstat(s, &d) < 0)
			goto fail;
		if(!incdir && (d.mode&CHDIR))
			continue;
		kids[j++] = d;
	}

	if(s)
		free(s);
	if(dentp){
		for(i=0; i<nkids; i++)
			free(dentp[i]);
		free(dentp);
	}

	*kidsp = kids;
	return j;

 fail:
	if(s)
		free(s);
	if(kids)
		free(kids);
	if(dentp){
		for(i=0; i<nkids; i++)
			free(dentp[i]);
		free(dentp);
	}

	return -1;
}
#endif

static void*
emalloc(long sz)
{
	void *v;

	v = malloc(sz);
	if(v == nil)
		sysfatal("malloc %lud fails\n", sz);
	memset(v, 0, sz);
	return v;
}

static void*
erealloc(void *p, long sz)
{
	p = realloc(p, sz);
	if(p == nil)
		sysfatal("realloc %lud fails\n", sz);
	return p;
}

static char*
estrdup(char *s)
{
	s = strdup(s);
	if(s == nil)
		sysfatal("strdup fails\n");
	return s;
}

static char*
estrpath(char *dir, char *name)
{
	char *u;

	if(dir[strlen(dir)-1] == '/')
		return estrstrdup(dir, name);
	
	u = emalloc(strlen(dir)+1+strlen(name)+1);
	strcpy(u, dir);
	strcat(u, "/");
	strcat(u, name);
	return u;
}

static char*
estrstrdup(char *s, char *t)
{
	char *u;

	u = emalloc(strlen(s)+strlen(t)+1);
	strcpy(u, s);
	strcat(u, t);
	return u;
}

static int
ispathprefix(char *x, char *xslashy)
{
	int l;

	l = strlen(x);
	return strncmp(x, xslashy, l)==0 && (xslashy[l]=='/' || xslashy[l]=='\0' || (l>0 && x[l-1]=='/'));
}

static int
proto_getfields(char *str, char **args, int max, int mflag, char *set)
{
	Rune r;
	int nr, intok, narg, esc, quote;
	
	if(max <= 0)
		return 0;

	narg = 0;
	args[narg] = str;
	if(!mflag)
		narg++;
	intok = esc = quote = 0;
	for(;; str += nr) {
		nr = chartorune(&r, str);
		switch (r) {
		    case 0:
			goto done;
			break;
		    case (Rune) '\\':
			if(esc)
				esc = 0;
			else{
				esc = 1;
				memcpy(str, str+1, strlen(str));
			}
			break;
		    case (Rune) '"':
			if(!esc && quote)
				quote = 0;
			else
				quote = 1;
			if(!esc)
				memcpy(str, str+1, strlen(str));
		    default:
			esc = 0;
		}

		if(!esc && !quote && utfrune(set, r)) {
			if(narg >= max)
				break;
			*str = 0;
			intok = 0;
			args[narg] = str + nr;
			if(!mflag)
				narg++;
		} else {
			if(!intok && mflag)
				narg++;
			intok = 1;
		}
	}
done:
	return narg;
}

static int
proto_tokenize(char *str, char **args, int max)
{
	return proto_getfields(str, args, max, 1, " \t\n\r");
}
