#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include <libsec.h>

#include "iso9660.h"

typedef struct	Id Id;

enum {
	IDTABSZ = 127,
	Unknown = 1,
	Plan9,
	Unix
};

struct Id {
	long	id;
	char*	name;
	Id*	idlink;
	Id*	nmlink;
};

struct Idtab {
	Id*	id[IDTABSZ];
	Id*	name[IDTABSZ];
};

Idtab	*usertab;
Idtab	*grouptab;
static Idtab	*nousertab;
static Idtab	*nogrouptab;

static int	isnumber(char*);
static Id*	lookid(Idtab*, long);
static Id*	lookname(Idtab*, char*);
static void	putid(Idtab*, Id*);
static Id*	mkid(char *name, long id);
static int	sniff(Biobuf*);

static void	parseplan9(Idtab*, int*, char*, int, char*);
static void	parseunix(Idtab*, int*, char*, int, char*);

long
altgetnumid(Idtab *p, char *s)
{
	Id *id;

	if(p == nil)
		return ~0;
	if((id = lookname(p, s)) == nil)
		return ~0;
	return id->id;
}

char*
altgetstrid(Idtab *p, long n)
{
	Id *id;

	if(p == nil)
		return nil;
	if((id = lookid(p, n)) == nil)
		return nil;
	return estrdup(id->name);
}

char*
altnouserstr(long uid)
{
	Id *id;

	if(nousertab == nil)
		nousertab = mkidtab();
	if(lookid(nousertab, uid) == nil){
		fprint(2, "warning: unknown uid '%ld'\n", uid);
		if((id = mkid("unknown", uid)) != nil)
			putid(nousertab, id);
	}
	return estrdup("root");
}

char*
altnogroupstr(long gid)
{
	Id *id;

	if(nogrouptab == nil)
		nogrouptab = mkidtab();
	if(lookid(nogrouptab, gid) == nil){
		fprint(2, "warning: unknown gid '%ld'\n", gid);
		if((id = mkid("unknown", gid)) != nil)
			putid(nogrouptab, id);
	}
	return estrdup("wheel");
}

void
dumptab(Idtab *p)
{
	int i;
	Id *cur;

	for(i=0; i<IDTABSZ; i++)
		for(cur=p->id[i]; cur != nil; cur=cur->idlink)
			print("%s:*:%ld\n", cur->name, cur->id);
	return;
}

Idtab*
mkidtab(void)
{
	Idtab *p;

	p = emalloc(sizeof *p);
	return p;
}

Idtab*
openidtab(char *fname)
{
	char *s;
	int n, warned;
	Idtab *p;
	Biobuf *b;
	void (*parsefn)(Idtab*, int*, char*, int, char*);

	p = nil;
	warned = 0;

	if(fname == nil)
		return nil;

	if((b = Bopen(fname, OREAD)) == nil){
		fprint(2, "warning: unable to open %s: %r\n", fname);
		return nil;
	}
	switch(sniff(b)){
	default:
		abort();

	case Unknown:
		werrstr("%s: unknown file format", fname);
		goto Out;

	case Plan9:
		parsefn = parseplan9;
		break;

	case Unix:
		parsefn = parseunix;
		break;
	}
	if(Bseek(b, 0, 0) != 0)
		goto Out;
	p = mkidtab();
	for(n=0; (s = Brdstr(b, '\n', 1)) != nil; n++){
		/* skip blank lines and lines starting with '#' */
		while(isspace(*s))
			s++;
		if(*s != '#' && *s != '\0')
			parsefn(p, &warned, fname, n, s);
		free(s);
	}

 Out:
	Bterm(b);
	return p;
}

static Id*
lookname(Idtab *p, char *s)
{
	uint n;
	Id *cur;

	n = hash(s) % IDTABSZ;
	for(cur = p->name[n]; cur != nil; cur = cur->nmlink)
		if(strcmp(s, cur->name) == 0)
			break;
	return cur;
}

static Id*
lookid(Idtab *p, long id)
{
	uint n;
	char buf[64];
	Id *cur;

	snprint(buf, sizeof buf, "%ld", id);
	n = hash(buf) % IDTABSZ;
	for(cur = p->id[n]; cur != nil; cur = cur->idlink)
		if(id == cur->id)
			break;
	return cur;
}

static Id*
mkid(char *name, long id)
{
	Id *p;

	p = emalloc(sizeof *p);
	p->id = id;
	p->name = atom(name);

	return p;
}

static void
putid(Idtab *p, Id *id)
{
	char buf[64];
	uint nid, nname;

	snprint(buf, sizeof buf, "%ld", id->id);
	nid = hash(buf) % IDTABSZ;
	nname = hash(id->name) % IDTABSZ;

	id->idlink = p->id[nid];
	p->id[nid] = id;
	id->nmlink = p->name[nname];
	p->name[nname] = id;

	return;
}

/*
 * /adm/users is
 *	id:user/group:head member:other members
 *
 * /etc/{passwd,group}
 *	name:x:nn:other stuff
 */

static void
parseplan9(Idtab *p, int *warned, char *fname, int n, char *s)
{
	USED(p); USED(fname); USED(n); USED(s);

	if(*warned == 0)
		fprint(2, "warning: plan9 uid/gid parser not implemented\n");
	*warned = 1;
	return;
}

static void
parseunix(Idtab *p, int *warned, char *fname, int n, char *s)
{
	int nf;
	long id;
	char *e, *f[16];
	Id *nid;

	memset(f, 0, sizeof(f));
	if((nf = getfields(s, f, nelem(f), 1, ":")) < 3)
		goto Warn;
	id = strtol(f[2], &e, 10);
	if(e == f[2] || *e != '\0')
		goto Warn;
	if((nid = mkid(f[0], id)) == nil)
		sysfatal("mkid: %r");
	putid(p, nid);
	return;

 Warn:
	if(*warned == 0)
		fprint(2, "warning: unix id file '%s', "
		       "line %d botched\n", fname, n);
	*warned = 1;
	return;
}

static int
sniff(Biobuf *b)
{
	int nf, rv;
	char *p, *f[16];

	p = nil;
	if(Bseek(b, 0, 0) != 0)
		return Unknown;
	if((p = Brdstr(b, '\n', 1)) == nil)
		return Unknown;
	if((nf = getfields(p, f, nelem(f), 1, ":")) < 4){
		rv = Unknown;
		goto Out;
	}
	if(isnumber(f[0]) && !isnumber(f[2])){
		rv = Plan9;
		goto Out;
	}
	if(!isnumber(f[0]) && isnumber(f[2])){
		rv = Unix;
		goto Out;
	}
	rv = Unknown;

 Out:
	free(p);
	return rv;
}

static int
isnumber(char *s)
{
	char *q;

	strtol(s, &q, 10);
	return *q == '\0';
}
